var FFpreferences = null;

var skip1proxyprompt = false;

var flagfoxstrings = null; // localized strings
var countrynames = null;   // localized country names

var FlagfoxPollLoopID;

function Flagfox_startup()
{
    try
    {
        if (FlagfoxPollLoopID !== undefined)
            throw "Flagfox is already loaded!";

        countrynames = document.getElementById("flagfox-countrynames");
        if (!countrynames || !countrynames.stringBundle.getSimpleEnumeration().hasMoreElements())
            throw "Could not load country names";

        flagfoxstrings = document.getElementById("flagfox-localizedstrings");
        if (!flagfoxstrings || !flagfoxstrings.stringBundle.getSimpleEnumeration().hasMoreElements())
            throw "Could not load localized strings";

        flagState.icon = document.getElementById("flagfox-icon");
        if (!flagState.icon)
            throw "Could not find Flagfox icon";

        FFpreferences = Components.classes["@mozilla.org/preferences-service;1"]
                                  .getService(Components.interfaces.nsIPrefBranch);

        Flagfox_loadIPDB();               // Load IPDB from "profileDir/extensionDir/db.dat"
        Flagfox_dnsHandler.init();        // Fetch the services we need for DNS
        Flagfox_cleanupOldPrefs();        // Clear any pref entries from old versions
        Flagfox_setIconPos();             // Place flag icon in the status bar
        Flagfox_prefObserver.register();  // Listen for preferences changes
    }
    catch(e)
    {
        window.removeEventListener("load", Flagfox_startup, false);
        window.removeEventListener("unload", Flagfox_shutdown, false);
        Flagfox_shutdown();

        Flagfox_error(true,"Flagfox failed to load!",e);
        return;
    }

    // Start polling loop
    FlagfoxPollLoopID = setInterval(Flagfox_pollWindowURL, 250);
}

function Flagfox_shutdown()
{
    clearInterval(FlagfoxPollLoopID);
    Flagfox_dnsHandler.reset();
    Flagfox_prefObserver.unregister();
    Flagfox_closeIPDB();
}

var flagState =
{
    url : "init",
    host : "",
    ip : "",
    country : null,
    icon : null,

    newURL : function(newurl)
    {
        this.url = newurl;
        this.host = "";
        this.ip = "";
        this.country = null;
    },

    toString : function()
    {
        var server;
        if (this.host != this.ip)
        {
            switch (textdirection)  // via global.dtd
            {
                default:
                case "ltr":
                    server = this.host + " (" + this.ip + ")";
                    break;
                case "rtl":
                    server = "(" + this.ip + ") " + this.host;  // Can't do "host (ip)" as that turns into "(host (ip"...
                    break;  // This outputs as "host (ip)"; there is no way to output "(ip) host" in RTL locales.
            }
        }
        else
        {
            server = this.ip;  // No hostname; only IP
        }
        return flagfoxstrings.getFormattedString( "tooltip", [server,this.country[1]] );
    }
};

var Flagfox_prefObserver =
{
    branch : null,  // the pref branch we're observing

    register : function()
    {
        var prefService = Components.classes["@mozilla.org/preferences-service;1"]
                                    .getService(Components.interfaces.nsIPrefService);

        this.branch = prefService.getBranch("flagfox.");
        this.branch.QueryInterface(Components.interfaces.nsIPrefBranch2);
        this.branch.addObserver("", this, false);
    },

    unregister : function()
    {
        if (this.branch != null)
            this.branch.removeObserver("", this);
    },

    observe : function(aSubject, aTopic, aData)
    {
        if  (aTopic != "nsPref:changed") return;
        switch (aData)
        {
            case "usealticons":
                Flagfox_setIconSrc();
                break;
            case "iconposition":
                Flagfox_setIconPos();
                break;
        }
    }
};

function Flagfox_setIconPos()
{
    try
    {
        var statusbar = document.getElementById("status-bar");
        var flagbox = document.getElementById("flagfox-statusbarpanel");
        var element;
        switch (FFpreferences.getCharPref("flagfox.iconposition"))  // Positions reversed in right-to-left locales
        {
            case "L":
                element = document.getElementById("statusbar-display");
                break;
            case "R":
                element = document.getElementById("statusbar-progresspanel").nextSibling;
                break;
            case "RM":
                element = statusbar.lastChild.nextSibling;  // Before after last == after  ;)
                break;
            default:
                throw "Invalid Flagfox icon position";
        }
        statusbar.insertBefore(flagbox,element);  // There is no insertAfter()
    }
    catch(e) { Flagfox_error(true,"Failed to set icon position",e); }
}

function Flagfox_setIconSrc()  // Applies to looked-up IPs only; others are done in the protocol switch
{
    if (flagState.country == null)
        return;  // Flag is for local file or unknown site

    var file;
    switch (flagState.country[0])
    {
        case '-A':  case '-B':  case '-C':
            file = "special/privateip";
            break;
        case '-L':
            file = "special/localhost";
            break;
        case 'A1':  case 'A2':
            file = "special/anonymous";
            break;
        default:
            var set = FFpreferences.getBoolPref("flagfox.usealticons") ? "flagset2/" : "flagset1/";
            file = set + flagState.country[0].toLowerCase();
            break;
    }
    flagState.icon.src = "chrome://flagfox/content/icons/" + file + ".png";
}

function Flagfox_pollWindowURL()
{
    try
    {
        if (flagState.url != window.content.document.location.href)
        {
            flagState.newURL(window.content.document.location.href);
            Flagfox_dnsHandler.reset();  // If we've changed pages before completing a lookup, then abort the old request and use this new one

            switch (window.content.document.location.protocol)
            {
                case "file:":
                    flagState.icon.src = "chrome://flagfox/content/icons/special/localfile.png";
                    flagState.icon.tooltipText = flagfoxstrings.getString("localfile");
                    return;  // Done

                case "about:":  case "chrome:":  case "resource:":  case "data:":
                    if (window.content.document.URL == "about:blank")
                    {   // Blank page gets its own icon and tooltip
                        flagState.icon.src = "chrome://flagfox/content/icons/special/blank.png";
                        flagState.icon.tooltipText = flagfoxstrings.getString("blankpage");
                    }
                    else
                    {
                        flagState.icon.src = "chrome://flagfox/content/icons/special/resource.png";
                        flagState.icon.tooltipText = flagfoxstrings.getString("localfile") + " (" + window.content.document.location.protocol + "//)";
                    }
                    return;  // Done

                default:
                    flagState.icon.src = "chrome://flagfox/content/icons/special/unknown.png";
                    flagState.icon.tooltipText = flagfoxstrings.getString("unknownsite");
                    break;  // Consider unknown until we can look it up
            }

            flagState.host = window.content.document.location.hostname;
            flagState.host = flagState.host.cropTrailingChar(".");  // Get rid of root dot, if it's there
            if (flagState.host == "")
                return;

            if (Flagfox_proxyCheck())
            {
                flagState.icon.src = "chrome://flagfox/content/icons/special/anonymous.png";
                flagState.icon.tooltipText += " (" + flagfoxstrings.getString("proxyprompttitle") + ")";  // correct in LTR & RTL
                return;  // Proxy in use; can't do a DNS lookup
            }

            // Idealy just hitting the DNS cache here
            Flagfox_dnsHandler.resolveHost();
            // Calls back to Flagfox_dnsHandler.onLookupComplete()
        }
    }
    catch (e) { Flagfox_error(false,null,e); }
}

var Flagfox_dnsHandler =
{
    dns : null,      // DNS service
    ios : null,      // I/O service
    proxy : null,    // proxy service
    thread : null,   // this thread; onLookupComplete() must be called in this thread
    request : null,  // current DNS request in progress; cancel with this.reset()

    init : function()
    {

        this.dns = Components.classes["@mozilla.org/network/dns-service;1"]
                             .getService(Components.interfaces.nsIDNSService);

        this.ios = Components.classes["@mozilla.org/network/io-service;1"]
                             .getService(Components.interfaces.nsIIOService);

        this.proxy = Components.classes["@mozilla.org/network/protocol-proxy-service;1"]
                               .getService(Components.interfaces.nsIProtocolProxyService);

        var EQS = Components.classes["@mozilla.org/event-queue-service;1"];
        if (EQS)  // Firefox 1.5 - 2.0
        {
            EQS = EQS.getService(Components.interfaces.nsIEventQueueService);
            this.thread = EQS.getSpecialEventQueue(EQS.CURRENT_THREAD_EVENT_QUEUE);
        }
        else      // Firefox 3.0+
        {
            this.thread = Components.classes["@mozilla.org/thread-manager;1"]
                                    .getService(Components.interfaces.nsIThreadManager)
                                    .currentThread;
        }
        if (!this.thread)
            throw "Could not fetch current thread for DNS handler";
    },

    reset : function()
    {
        if (this.request != null)
        {
            this.request.cancel(Components.results.NS_ERROR_ABORT);
            this.request = null;
        }
    },

    detectProxy : function()  // Returns true if the current URL is set to have its DNS lookup proxied via SOCKS
    {
        try
        {
            var uri = this.ios.newURI(flagState.url, null, null);
            var proxyinfo = this.proxy.resolve(uri, 0);  // Finds proxy (shouldn't block thread; we already did this lookup to load the page)
            return (proxyinfo != null) && (proxyinfo.flags & proxyinfo.TRANSPARENT_PROXY_RESOLVES_HOST);
            // "network.proxy.socks_remote_dns" pref must be set to true for Firefox to set TRANSPARENT_PROXY_RESOLVES_HOST flag when applicable
        }
        catch (e)
        {
            Flagfox_error(true,"Proxy check failed",e);
            throw "Lookup abort due to proxy check fail";
        }
    },

    resolveHost : function()
    {
        this.request = this.dns.asyncResolve(flagState.host, 0, this, this.thread);
    },

    onLookupComplete : function(nsrequest, nsrecord, status)
    {
        this.request = null;
        if (status != 0 || !nsrecord || !nsrecord.hasMore())
            return;  // IP not found in DNS  (status=0x804B001E for DNS resolution fail)

        flagState.ip = nsrecord.getNextAddrAsString();
        flagState.country = Flagfox_lookupIP(flagState.ip);
        if (flagState.country == null)
            return;  // IP is not in DB

        // Set the image
        Flagfox_setIconSrc();
        // Set the tooltip
        flagState.icon.tooltipText = flagState.toString();
    }
};

function Flagfox_proxyCheck()  // Return true to abort DNS hit (false allows local DNS request if not cached)
{
    switch (FFpreferences.getIntPref("flagfox.dnswithproxy"))
    {
        case -1:  // Deny
            return Flagfox_dnsHandler.detectProxy();

        default:  // Error
            FFpreferences.setIntPref("flagfox.dnswithproxy",0);
            Flagfox_error(true,"Invalid DNS setting; preference has been reset",null);
            // Reset and fallthrough to prompt

        case 0:   // Prompt
            if (skip1proxyprompt)  // Skips duplicate prompts when pressing the FAQ button
            {
                skip1proxyprompt = false;
                return true;  // Deny
            }
            if (Flagfox_dnsHandler.detectProxy())
            {
                const prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                                          .getService(Components.interfaces.nsIPromptService);
                const flags = prompts.BUTTON_POS_2 * prompts.BUTTON_TITLE_YES +
                              prompts.BUTTON_POS_1 * prompts.BUTTON_TITLE_IS_STRING +
                              prompts.BUTTON_POS_0 * prompts.BUTTON_TITLE_NO +
                              prompts.BUTTON_POS_0_DEFAULT +
                              prompts.BUTTON_DELAY_ENABLE;  // Stays greyed out for a moment to force users to think
                var check = {value: false};  // Default the checkbox to false
                var button = prompts.confirmEx( null,
                                                flagfoxstrings.getString("proxyprompttitle"),
                                                flagfoxstrings.getString("proxypromptmessge"),
                                                flags,
                                                "", "FAQ", "",
                                                flagfoxstrings.getString("proxypromptchecklabel"),
                                                check );

                if (button == 1)  // Open FAQ page (English forums, so I'm not going to translate the button label)
                {
                    skip1proxyprompt = true;
                    const thisBrowser = window.getBrowser();  // Must open in this window; vars won't persist to a new window
                    thisBrowser.selectedTab = thisBrowser.addTab("http://flagfox.servehttp.com/forum/viewtopic.php?f=3&t=20",null,null);
                    return true;  // Deny
                }
                var deny = (button==0);
                if (check.value)
                    FFpreferences.setIntPref( "flagfox.dnswithproxy", deny?-1:1 );
                return deny;
            }
            // else fallthrough and allow

        case 1:   // Allow
            return false;
    }
}

function Flagfox_cleanupOldPrefs()
{
    const OLD_PREFS = ["flagsPrefs","posPrefs","linkPrefs","tabPrefs","allowdnslookup"];
    for (var p in OLD_PREFS)
    {
        var entry = "flagfox." + OLD_PREFS[p];
        if (FFpreferences.getPrefType(entry) != 0)
            FFpreferences.clearUserPref(entry);
    }
}

String.prototype.cropTrailingChar = function(character)
{
    return (this.charAt(this.length-1)==character) ? this.slice(0,this.length-1) : this.valueOf();
};
